/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/

#include "common.h"

#define PROGNAME "read"
#define VERSION "0.3"

#define DEBUG_
#define DONTREAD_

#if EXTRADEFNOTUSED
 #define READTEST
 #define READRANDOM
#endif

#ifdef READTEST
/*recursively walk through dirs*/
 #include "cl_array.h"
 #include "cl_string.h"
 #include "cl_time.h"
// #include "cl_hist.h"
#ifdef READRANDOM
/*randomly read files*/
 #include <gsl/gsl_math.h>
 #include <gsl/gsl_rng.h>
 #include <gsl/gsl_randist.h>
 #define SPEED
 #define FIBTEST
 //#define speedtest
#endif
#endif

#ifdef READRANDOM
#ifdef SPEED
#define DIRPREALLOC 35
#define FILEPREALLOC 15
struct st_file{
	char* name;
	unsigned int dir;
	unsigned long length;
	unsigned int visited;
#ifdef FIBTEST
	unsigned int fibblock;
#endif
};

/*
struct fileinfo {
	char *name;
	unsigned long length;
	unsigned int dir;
};

struct dirinfo {
	char* path;
	unsigned long f_no;
	unsigned long c_no;
	unsigned long d_no;
};
*/
#else
template class cl_Array<cl_String>;
struct st_file {
	cl_String name;
	unsigned int dir;
	unsigned long length;
	unsigned int visited;
};
template class cl_Array<st_file>;
#endif
#endif

int verbose=0;
int blocksize=BUFFERSIZE;
char *dummybuffer;

void usage(){
#ifdef READTEST
	printf("Usage of %s v%s (READTEST defined)\n",PROGNAME,VERSION);
#else
	printf("Usage of %s v%s\n",PROGNAME,VERSION);
#endif
	printf("\t -b blksize\t\tblocksize to use for buffered i/o\n");
	printf("\t -d dirname\t\tchange to directory \"dirname\"\n");
	printf("\t -h,-?\t\t\tthis text\n");
	printf("\t -v\t\t\tbe verbose\n");
#ifdef READRANDOM
#ifdef SPEED
	printf("\t -p maxd:maxf\t\tset # of maxdirs and maxfiles\n");
#endif
	printf("\t -n files\t\tset # files to random read\n");
	printf("\t -l kbytes\t\tset # kbytes to random read\n");
	printf("\t -y seed\t\tset random generator seed\n");
	printf("\t -z\t\t\tsync with unmount (don't use or edit agesync.sh)\n");
#endif
	exit(1);
}

#ifdef READTEST
cl_Timer timer(1);
#define Nr_ 20
cl_Timer timer2(Nr_);

/* just for curiosity */
unsigned long gcounterdirs=0;
unsigned long long gcounterfiles=0;
unsigned long gcounterslinks=0;
unsigned long gcounterelses=0;

unsigned long long gcountbytes=0;


#ifdef READRANDOM
#ifdef SPEED
unsigned long gmaxdirs=15000;
unsigned long long gmaxfiles=100*gmaxdirs;
char** dirs;
st_file* files;
#else
cl_Array<cl_String> dirs;
cl_Array<st_file> files;
#endif
unsigned long long gtotalfiles=0;
unsigned long gtotaldirs=0;
#endif

int fatal(char *mess)
{
	fprintf(stderr, "Error: %s\n", mess); 
	exit(1);
}

int rekdir_func_stat(const char* path, const struct stat *sfs)
{
#ifdef READRANDOM
#ifdef SPEED
	st_file* tmpfile;
#else
	st_file tmpfile;
	cl_String tmp(256);
#endif
#endif	

	if(S_ISDIR(sfs->st_mode)==1) {
#ifdef READRANDOM
#ifdef SPEED
		if(gcounterdirs==gmaxdirs){
			fprintf(stderr,"Error in rek_dir_func_stat: maximum dir number reached (%ld)\n",gmaxdirs);
			exit(1);
		}	

		if(strlen(path)+1>DIRPREALLOC) {
			delete[] dirs[gcounterdirs];
			dirs[gcounterdirs]=new char[strlen(path)+1];
		}
		strcpy(dirs[gcounterdirs],path);
#else
		tmp=path;
		dirs.append(&tmp,1);
#endif
		gcounterdirs++;
#endif
		if(verbose>2) printf("Working in directory: %s\n",path);
	}
	else if(S_ISREG(sfs->st_mode)==1)
	{
		long length=sfs->st_size;

		int i=0;
		long count=1024;
		while((i<Nr_)&&(count<length)){
			i++;
			count*=2;
		}
		
		if(verbose>=1) printf("Reading file \"%s\" done\n",path);
		timer2.timestamp();
#ifndef DONTREAD
		if(readfile(path,length,O_RDONLY)==-1) return -1;
#endif
#ifdef DEBUG
		double time=timer2.time_so_far();
		printf("%s: %e KB/sek\n",path,length/time/1024);
#endif
		timer2.get_delta_t(i,length);

		gcountbytes+=length;
#ifdef READRANDOM
#ifdef SPEED
		if(gcounterfiles==gmaxfiles){
			fprintf(stderr,"Error in rek_dir_func_stat: maximum files number reached (%llu)\n",gmaxfiles);
			exit(1);
		}	

		tmpfile=&files[gcounterfiles];
		//get file name;
		char* tmp=rindex(path,'/');
		int a=strlen(&tmp[1]);
		if(a+1>FILEPREALLOC) {
			delete[] tmpfile->name;
			tmpfile->name=new char[a+1];
		}
		strcpy(tmpfile->name,&tmp[1]);
		tmpfile->visited=0;
		tmpfile->length=length;
		//get dir entry
		int b=strlen(path)-a-1;
		a=gcounterdirs-1;
		while ((a>=0)&&(((int)strlen(dirs[a])!=b)||(strncmp(dirs[a],path,b)!=0))) a--; 
		tmpfile->dir=a;		
		//printf("%d: %s/%s == %s\n",tmpfile->dir,dirs[a],tmpfile->name,path);
#else
		tmp=path;
		//get file name;
		int a=tmp.length()-1;
		int b=a;
		while((tmp[a]!='/')&&(a>0)) a--;
		tmpfile.name=tmp.mid(a+2,b+1);
		tmp.cut(a,b+1);
		//get dir entry
		a=dirs.length()-1;
		while ((a>=0)&&(dirs[a]!=tmp)) a--; 
		tmpfile.dir=a;
		tmpfile.length=length;
		tmpfile.visited=0;
		files.append(&tmpfile,1);
#endif
		gcounterfiles++;
#endif
	}
	else if(S_ISLNK(sfs->st_mode)==1){
		gcounterslinks++;
		if(verbose>2) printf("Omitting slink: %s\n",path);
	}
	else {
		gcounterelses++;
		if(verbose>2) printf("Omitting: %s\n",path);
	}
	return 0;
}

#ifdef READRANDOM
gsl_rng *r;  /* global generator */

//bad hack
cl_String tmpdir="";
int flag=0;

void clsync(){
	if((flag!=0)&&(tmpdir.length()>0))
	 if(chdir("/root/bin")==0){
		cl_String cmd="agesync.sh ";
		cmd+=tmpdir;
		//cout << cmd << " ---- " << endl; 
		int ret=system((char *)cmd);
		if(ret!=0){
			fprintf(stderr,"Error call agesync.sh\n");
			exit(1);
		}		
		chdir((char*)tmpdir);
	}
}

int randomreadtest(long readtestfiles=-1,unsigned long readbytes=0)
{
	cl_String file(1024);
	file.clear();

	//unsigned long long which=0;
	int which=0;
	long length=0;
	if((readtestfiles==-1)&&(readbytes==0)) readtestfiles=(long)gtotalfiles/10;	

	while(1){
		which=gsl_rng_uniform_int(r,gtotalfiles);
		if(files[which].visited % 3!=0) {
			files[which].visited+=1;
			continue;
		}
		if(files[which].visited>0){
			clsync(); /* to be sure file is not in cache */
			for(unsigned int i=0;i<gtotalfiles;i++) files[i].visited=0;
		}
		length=files[which].length;
		file=dirs[files[which].dir];
		file+="/";
		file+=files[which].name;
		files[which].visited+=1;
		
		int i=0;
		long count=1024;
		while((i<Nr_)&&(count<length)){
			i++;
			count*=2;
		}

		if(verbose>=1) printf("Reading file \"%s\" done\n",(char *)file);
		timer2.timestamp();
#ifndef DONTREAD
		if(readfile((char*)file,length,O_RDONLY)==-1) fatal("Error in randomreadtest");
#endif
#ifdef DEBUG
		double time=timer2.time_so_far();
		printf("%s: %e KB/sek\n",(char*)file,length/time/1024);
#endif
		timer2.get_delta_t(i,length);

		gcountbytes+=length;
		gcounterfiles++;
		if((readtestfiles!=0)&&(gcounterfiles>=(unsigned long)readtestfiles)) break;
		if((readbytes>0)&&((float)gcountbytes/1024>=readbytes)) break;
	}
	return 0;
}
#ifdef FIBTEST
int getandsortfibs()
{
	cl_String file(1024);
	file.clear();

	//get fibs
	int fd;
	long fib_res=0,fib_blk=0;

	for(unsigned long long which=0;which<gtotalfiles;which++) {

		file=dirs[files[which].dir];
		file+="/";
		file+=files[which].name;

		files[which].visited=which;
		files[which].fibblock=0;

		fd=open((char*)file,O_RDONLY);
		if(fd==-1){
			fprintf(stderr,"Error opening \"%s\"\n",(char*)file);
			continue;
		}

		fib_blk=0;
		fib_res=ioctl(fd,FIBMAP,&fib_blk);
		//cout << file << ": " << fib_res << " " << fib_blk << endl;		

		close(fd);
		if(fib_res==-1)
		{
			fprintf(stderr,"Error fipmaping \"%s\"\n",(char*)file);
			continue;
		}

		files[which].fibblock=fib_blk;
	}

	//sort fibs
	int weiter=1;
	int k=gtotalfiles;
	long tmp1,tmp2;
	while(weiter==1)
	{
		weiter=0;
		for(int j=1;j<k;j++)
		{
			if(files[j].fibblock<files[j-1].fibblock)
			{
			 	weiter=1;
				tmp1=files[j-1].fibblock;
				tmp2=files[j-1].visited;
				
				files[j-1].fibblock=files[j].fibblock;
				files[j-1].visited=files[j].visited;
				files[j].fibblock=tmp1;
				files[j].visited=tmp2;
			}	
		}
	}
/*
	for(unsigned long long which=0;which<gtotalfiles;which++) 
	{	
		printf("%5llu: %s %u %u\n",which,files[files[which].visited].name,files[files[which].visited].dir,files[which].fibblock);
	}
*/
	return 0;
}

int fibreadtest()
{
	cl_String file(1024);
	file.clear();
	long length=-1;
		
	for(unsigned long long which=0;which<gtotalfiles;which++) 
	{
		//printf("%5llu: %s %u %u\n",which,files[files[which].visited].name,files[files[which].visited].dir,files[which].fibblock);
		if(files[which].fibblock==0) continue; //omitt tails
		
		file=dirs[files[files[which].visited].dir];
		file+="/";
		file+=files[files[which].visited].name;
	
		length=files[files[which].visited].length;
		int i=0;
		long count=1024;
		while((i<Nr_)&&(count<length)){
			i++;
			count*=2;
		}

		timer2.timestamp();
#ifndef DONTREAD
		if(readfile((char*)file,-1,O_RDONLY)==-1) fatal("Error in fibreadtest");
#endif
#ifdef DEBUG
		double time=timer2.time_so_far();
		printf("%s: %e KB/sek\n",(char*)file,length/time/1024);
#endif
		timer2.get_delta_t(i,length);
		if(verbose>=1) printf("Reading file \"%s\" done\n",(char *)file);

		gcountbytes+=length;
		gcounterfiles++;
	}
	return 0;
}

#endif

#endif
#endif


int main(int argc,char **argv)
{
#ifdef READRANDOM
	long readfnumber=-1;
	long readbytes=-1; /*KB*/
	long seed=0;
#endif
			
	/*parse options*/
	int temp;
	char ch;
#ifdef READRANDOM
#ifdef SPEED
	while((ch=getopt(argc,argv,"b:d:n:p:l:h?vy:z"))!=EOF){
#else
	while((ch=getopt(argc,argv,"b:d:n:l:h?vy:z"))!=EOF){
#endif
#else
	while((ch=getopt(argc,argv,"b:d:h?v"))!=EOF){
#endif
		switch(ch)	{	
			case 'b':
				temp=atoi(optarg);
				if(temp<1)
				{
					fprintf(stderr, "Can't set blocksize to \"%s\".\n", optarg);
					usage();
				} else blocksize=temp;
				break;
			case 'd':
				if(chdir(optarg))
				{
					fprintf(stderr, "Can't change to directory \"%s\".\n", optarg);
					usage();
				} 
#ifdef READRANDOM
					else tmpdir=optarg;
#endif
				break;
#ifdef READRANDOM
#ifdef SPEED
			case 'p':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<2)){
					//printf("%d-%s\n",++i,opttok);
					temp=atoi(opttok);
					if(temp<0)
					{
						fprintf(stderr, "Can't set directory counter to \"%s\".\n", optarg);
						usage();
					} 
					else if(i==1) gmaxdirs=temp;
					else { 
						gmaxfiles=temp;
						if(gmaxfiles<gmaxdirs) gmaxfiles=100*gmaxdirs;
					}
					opttok=strtok(NULL,":");
				}
				break;}
#endif
			case 'l':
				temp=atoi(optarg);
				if(temp<-1)
				{
					fprintf(stderr, "Can't set readbytes to \"%s\".\n", optarg);
					usage();
				} else readbytes=temp;
				break;
			case 'n':
				temp=atoi(optarg);
				if(temp<-1)
				{
					fprintf(stderr, "Can't set readnumber to \"%s\".\n", optarg);
					usage();
				} else readfnumber=temp;
				break;
			case 'z': flag=1; break;
#endif
			case '?':
			case 'h': usage(); break;
			case 'v': verbose++; break;
#ifdef READRANDOM
			case 'y':
				seed=atol(optarg);
				break;
#endif
			default: usage(); break;
		}
	}

	dummybuffer=new char[blocksize];
	memset(dummybuffer,31,blocksize);

	clsync();

#ifdef speedtest
	#define bn_ 100000
#ifdef SPEED
	st_file* tmpfile;
	files=new st_file[bn_];
	sleep(1);
	for(int i=0;i<bn_;i++){
		tmpfile=new st_file;
		tmpfile->name=new char (12);//"constantin";
		tmpfile->visited=1;
		tmpfile->length=i;
		files[i]=*tmpfile;
		if(i % 10000 == 0) {
			printf("%d\n",i);
			sleep(1);
		}
	}	
#else
	st_file tmpfile;
	tmpfile.name="constantin";
	tmpfile.visited=1;
	files.init(bn_);
	sleep(1);
	for(int i=0;i<bn_;i++){
		tmpfile.length=i;
		files.append(&tmpfile,1);
		files[i]=tmpfile;
		if(i % 10000 == 0) {
			printf("%d %ld\n",i,tmpfile.length);
			sleep(1);
		}
	}	
#endif
	for(int i=0;i<bn_;i++) printf("%d %ld\n",i,files[i].length);
	sleep(5);
	exit(1);
#endif

#ifdef READTEST
#ifdef READRANDOM
#ifdef SPEED
	dirs=new char*[gmaxdirs];
	for(unsigned long i=0;i<gmaxdirs;i++) dirs[i]=new char[DIRPREALLOC];
	files=new st_file[gmaxfiles];
	for(unsigned long long i=0;i<gmaxfiles;i++) files[i].name=new char[FILEPREALLOC];
/*
	printf("%d %d\n",gmaxdirs,gmaxfiles);
	dirinfo *dirs2=new dirinfo[gmaxdirs];
	fileinfo *files2=new fileinfo[gmaxfiles];
	for(unsigned long i=0;i<gmaxdirs;i++) dirs2[i].path=new char[DIRPREALLOC];
	for(unsigned long long i=0;i<gmaxfiles;i++) files2[i].name=new char[FILEPREALLOC];
	
	sleep(5);
	exit(1);
*/
#else
	dirs.init(10000);
	files.init(100000);
	dirs.clear();
	files.clear();
#endif
#endif
	timer.initialize();
	timer.timestamp();

	for(;optind<argc;optind++){
		if (walk_through_dirs(argv[optind],rekdir_func_stat)==-1) fatal("ret_walk_through_dirs failed...");
	}
#else
	for(;optind<argc;optind++){
		if(verbose==1) printf("Reading file \"%s\" done\n",argv[optind]);
#ifndef DONTREAD
		if(readfile(argv[optind],-1)==-1) exit(1);
#endif
	}
#endif

#ifdef READTEST
	double time=timer.time_so_far();
	timer.get_delta_t(0,0);
	printf("Total: %llu files (%llu Bytes in %e) in %lu dirs\n with average of %e KB/sek\n",gcounterfiles,gcountbytes,time,gcounterdirs,(double)gcountbytes/1024/time);
	for(int i=0;i<Nr_;i++) timer2.print_stat(i);

#ifdef READRANDOM
#ifdef DEBUG
#ifdef SPEED
	for(unsigned int i=0;i<gcounterdirs;i++) cout << dirs[i] << endl;
	for(unsigned int i=0;i<gcounterfiles;i++) cout << files[i].name << " " << files[i].length << " " << files[i].dir << endl;
#else
	for(unsigned int i=0;i<dirs.length();i++) cout << dirs.at(i) << endl;
	for(unsigned int i=0;i<files.length();i++) cout << files.at(i).name << " " << files.at(i).length << " " << files.at(i).dir << endl;
#endif
#endif

	clsync();

	gsl_rng_default_seed=seed;
	r=gsl_rng_alloc(gsl_rng_default);
	gtotalfiles=gcounterfiles;
	gtotaldirs=gcounterdirs;
	gcounterfiles=0;
	gcountbytes=0;

	timer.initialize();
	timer2.initialize();
	timer.timestamp();

	randomreadtest(readfnumber,readbytes);

	time=timer.time_so_far();
	if(gcountbytes==0) time=1; /*prevent floating error*/
	timer.get_delta_t(0,0);
	printf("Random Total: %llu files (%llu Bytes in %e) in %lu dirs\n with average of %e KB/sek\n",gcounterfiles,gcountbytes,time,gcounterdirs,(double)gcountbytes/1024/time);
	for(int i=0;i<Nr_;i++) timer2.print_stat(i);

#ifdef FIBTEST
	getandsortfibs();
	
	clsync();

	timer.initialize();
	timer2.initialize();
	timer.timestamp();

	gcounterfiles=0;
	gcountbytes=0;

	fibreadtest();

	time=timer.time_so_far();
	if(gcountbytes==0) time=1; /*prevent floating error*/
	timer.get_delta_t(0,0);
	printf("Fibread Total: %llu files (%llu Bytes in %e) in %lu dirs\n with average of %e KB/sek\n",gcounterfiles,gcountbytes,time,gcounterdirs,(double)gcountbytes/1024/time);
	for(int i=0;i<Nr_;i++) timer2.print_stat(i);

#endif
#endif

#endif
	delete[] dummybuffer;
}

